import numpy as np

class AADUtils:
    """
    AADUtils class: contains utility functions used by the AAD library.
    """

    @staticmethod
    def align_lists(a, b):
        """
        Returns the lists or np.ndarray a, b in a tuple with their lengths aligned and zero-padded. 
        If a no-len is passed, then zero list

        Inputs
        -----------------
        a           : list
                    list of values
        b           : list
                    list of values

        Outputs
        ------------------
        out         : tuple
                    tuple containt the lists or np.ndarrays of a and b. If a no-len is passed, then zero list

        Examples
        -----------------
        >>> AADUtils.align_lists([1, 0, 1, 2], [0, 0, 5, 0])
        ([1, 0, 1, 2], [0, 0, 5, 0])
        >>> AADUtils.align_lists([1], [0, 0, 5, 0])
        (array([1, 0, 0, 0]), [0, 0, 5, 0])
        >>> AADUtils.align_lists([1, 3], [2])
        ([1, 3], array([2, 0]))
        >>> AADUtils.align_lists([1, 3], 9) 
        ([1, 3], array([0, 0]))
        """
        la, lb = len(a) if hasattr(a, '__len__') else 0, len(b) if hasattr(b, '__len__') else 0
        if la == lb:
            return a, b
        elif la > lb:
            return a, np.array([b[j] if j < lb else 0 for j in range(la)])
        else:
            return np.array([a[j] if j < la else 0 for j in range(lb)]), b



# use pip to install: pip install -r requirements.txt
import numpy as np
import math
import copy

from AADUtils import AADUtils

class AADVariable:
    """ 
    AAD VARIABLE CLASS: 
    PROVIDES OVERRIDEN OPERATORS FOR CORRECT OPERATIONS BETWEEN AAD VARIABLES WITH THEMSELVES AND WITH REAL NUMBERS 
    RETURNS AAD CLASS

    Support for multivariate: self.der now accepts an array or scalar.
    When passing in arrays, the positioning refers to the variable given, i.e.
    AADVariable(val=2.0, der=1.0)   -or- is one variable operation mode, or just "x"
    AADVariable(val=2.0, der=[1.0])
    AADVariable(val=1.5, der=[0.0, 1.0]) refers to "y"

    All calculations will expand the derivative list to the maximum required.
    Scalars or 1-variable case will be handled as an exceptional case, and will return a scalar when requested (the getter will shim this),
    and accept a scalar for the class constructor. However, internally, self.der is always a list. (hplin, 11/27/20)

    Vector-valued functions are not implemented here. Instead, since each element in a vector-valued function is independent,
    they can be handled as a list of AADVariables using an external component. AADVariable is a multivariate-input, scalar-output
    Dual number class in the forward mode.

    Example
    ------------------
    >>> x = AADVariable(1, [1,0])
    >>> y = AADVariable(2, [0,1])

    >>> z1 = x + y
    >>> z2 = z1**2
    >>> z3 = sin(z2)+cos(y)
    >>> z3.val
    -0.004028351305385813
    >>> z3.der
    array([-5.46678157, -6.376079  ])

    """
    def __init__(self, val, der=1.0, name=None):
        '''
        Inputs
        ------------
        val     : float
                value set for the AADVariable.
        der     : float
                derivat
                
        
        '''
        self.name = name
        self.val = val
        if val == 0: der = 0. # check if the value passing in is 0, if so than der must be 0 by definition.

        self.der = der        # this has hidden implications - see der.setter below for the expected behavior.
    
    @property
    def der(self):
        """
        Returns the self.der property. For compatibility, when there is only one variable tracked, it returns a scalar.
        For code that is explicitly aware of this and handles all derivatives as lists, use self._der directly.
        """

        return self._der if len(self._der) > 1 else self._der[0]

    @der.setter
    def der(self, new):
        """Set the derivative. Accepts either a list i.e. [1,2,3], np.array([1,2,3]) or scalar (i.e. 2)"""
        if isinstance(new, list): # accepts list, convert to np.ndarray internally
            self._der = np.array(new)
        elif isinstance(new, np.ndarray):
            self._der = new
        else:
            self._der = np.array([new])

    def jacobian(self):
        """Return the Jacobian (a scalar for a scalar 1-variable function, or a matrix/vector for multivariate)"""
        return self.der


    def __eq__(self, other):
        """
        Equality operator
        """
        if not isinstance(other, AADVariable):
            return False
        
        d1, d2 = AADUtils.align_lists(self._der, other._der)
        res = (self.val == other.val and np.array_equal(d1, d2))
        return res

    def __ne__(self, other):
        return not self.__eq__(other)

    def __neg__(self):
        '''
        OVERLOADING NEGATION OPERATOR IE -SELF
        '''
        new = AADVariable(-self.val, -self.der) 
        return new

    def __add__(self, other):
        '''
        OVERLOADING ADDITION OPERATOR IE SELF+OTHER
        '''
        new = AADVariable(self.val)

        if isinstance(other, AADVariable):
            sv, ov = self.val, other.val
            sd, od = AADUtils.align_lists(self._der, other._der)
        else: # Is scalar
            sv, ov = self.val, other
            sd, od = AADUtils.align_lists(self._der, 0)
        new.val = sv + ov
        new.der = sd + od

        return new

    def __mul__(self, other):
        '''
        OVERLOADING THE MULTIPLICATION OPERATOR IE SELF*OTHER
        '''
        new = AADVariable(self.val)

        # Is AADVariable?
        if isinstance(other, AADVariable):
            sv, ov = self.val, other.val
            sd, od = AADUtils.align_lists(self._der, other._der)
        else: # Is scalar
            sv, ov = self.val, other
            sd, od = AADUtils.align_lists(self._der, 0)

        ## Computation
        new.val = sv * ov
        new.der = sd * ov + od * sv

        return new

    def __rmul__(self, other):
        '''
        OVERLOADING REVERSE MULTIPLICATION OPERATOR IE OTHER*SELF
        '''
        return self * other
        
    def __radd__(self, other):
        '''
        OVERLOADING REVERSE ADDITION OPERATOR IE OTHER+SELF
        '''
        return self + other

    def __sub__(self, other):
        '''
        OVERLOADING SUBTRACTION WITH NEGATION AND ADDITION IE SELF-OTHER
        '''
        return self + (-other)
    
    def __rsub__(self, other):
        '''
        OVERLOADING REVERSE SUBTRACTION WITH NEGATION AND ADDITION IE OTHER-SELF
        '''
        return -self + other

    def __truediv__(self, other): 
        '''
        OVERLOADING DIVISION OPERATOR IE SELF/OTHER
        '''
        new = AADVariable(self.val)

        # Is AADVariable?
        if isinstance(other, AADVariable):
            sv, ov = self.val, other.val
            sd, od = AADUtils.align_lists(self._der, other._der)
        else: # Is scalar
            sv, ov = self.val, other
            sd, od = AADUtils.align_lists(self._der, 0)

        ## Computation
        # (f/g)' = (f'g - g'f)/g**2
        new.val = sv / ov
        new.der = (sd * ov - sv * od)/(ov**2)

        return new

    def __rtruediv__(self, other): 
        '''
        OVERLOADING REVERSE DIVISION OPERATOR IE OTHER/SELF
        '''

        ## Boilerplate: Create new target variable, padd multivariates
        new = AADVariable(self.val)

        # Is AADVariable?
        if isinstance(other, AADVariable):
            sv, ov = self.val, other.val
            sd, od = AADUtils.align_lists(self._der, other._der)
        else: # Is scalar
            sv, ov = self.val, other
            sd, od = AADUtils.align_lists(self._der, 0)
        ## Computation
        new.val = ov / sv
        new.der = (od * sv - sd * ov)/(sv**2)
        return new

    def __pow__(self, other):
        '''
        OVERLOADING POWER OPERATOR IE SELF**OTHER
        '''
        new = AADVariable(0.0, 0.0)
        try:
            new.val = self.val ** other.val
            new.der = (self.val ** (other.val - 1)) * (self.der * other.val + self.val * math.log(self.val) * other.der)
           #EVALUATING EDGE CASES
        except AttributeError: # just simple case of number...
            new.val = self.val ** other
            new.der = self.val ** (other - 1) * other * self.der
        return new

    def __rpow__(self, other):
        '''
        OVERLOADING REVERSE POWER OPERATOR IE OTHER**SELF
        '''
        new = AADVariable(0.0, 0.0)
        try:
            new.val = other.val ** self.val
           #EVALUATING EDGE CASES
        except AttributeError: # just "simple" case of number... other**self
            new.val = other ** self.val
            new.der = math.log(other) * other**(self.val) * self.der
        return new

    def __repr__(self):
        '''
        OVERLOADING REPR FUNCTION FOR CLEAN DISPLAY
        '''
        return "AADVariable fun = " + str(self.val) + ", der = " + str(self.der) 



# use pip to install: pip install -r requirements.txt
import numpy as np
import math
import copy
from AADUtils import AADUtils
from AADVariable import AADVariable

def exp(obj: AADVariable) -> AADVariable:
    """ 
    EXP OPERATOR:
    INPUT: REAL NUMBER OR AAD-VARIABLE
    RETURNS: AAD-VARIABLE TYPE
    """
    name = obj.name
    val = obj.val
    der = obj.der
    n_val = np.exp(val)
    n_der = der * np.exp(val)
    return AADVariable(n_val,n_der,name=name)

def log(obj: AADVariable) -> AADVariable:
    """
    LOG BASE E OPERATOR:
    INPUT: REAL NUMBER OR AAD-VARIABLE
    RETURNS: AAD-VARIABLE TYPE
    """
    name = obj.name
    val = obj.val
    der = obj.der
    n_val = np.log(val)
    n_der = der * 1/(val)
    return AADVariable(n_val,n_der,name=name)

def sin(obj: AADVariable) -> AADVariable:
    """
    SIN OPERATOR:
    INPUT: REAL NUMBER OR AAD-VARIABLE
    RETURNS: AAD-VARIABLE TYPE
    """
    name = obj.name
    val = obj.val
    der = obj.der
    n_val = np.sin(val)
    n_der = np.cos(val) * der
    return AADVariable(n_val,n_der,name=name)

def sinh(obj: AADVariable) -> AADVariable:
    """
    SINH OPERATOR:
    INPUT: REAL NUMBER OR AAD-VARIABLE
    RETURNS: AAD-VARIABLE TYPE
    """
    name = obj.name
    val = obj.val
    der = obj.der
    n_val = np.sinh(val)
    n_der = der * np.cosh(val)
    return AADVariable(n_val,n_der,name=name)

def cos(obj: AADVariable) -> AADVariable:
    """
    COS OPERATOR:
    INPUT: REAL NUMBER OR AAD-VARIABLE
    RETURNS: AAD-VARIABLE TYPE
    """
    name = obj.name
    val = obj.val
    der = obj.der
    n_val = np.cos(val)
    n_der = der * -np.sin(val)
    return AADVariable(n_val,n_der,name=name)

def cosh(obj: AADVariable) -> AADVariable:
    """
    SINH OPERATOR:
    INPUT: REAL NUMBER OR AAD-VARIABLE
    RETURNS: AAD-VARIABLE TYPE
    """
    name = obj.name
    val = obj.val
    der = obj.der
    n_val = np.cosh(val)
    n_der = der * np.sinh(val)
    return AADVariable(n_val,n_der,name=name)

def tan(obj: AADVariable) -> AADVariable:
    """
    TAN OPERATOR:
    INPUT: REAL NUMBER OR AAD-VARIABLE
    RETURNS: AAD-VARIABLE TYPE
    """
    name = obj.name
    val = obj.val
    der = obj.der
    n_val = np.tan(val)
    n_der = der * 1/(np.cos(val)**2)
    return AADVariable(n_val,n_der,name=name)

def tanh(obj: AADVariable) -> AADVariable:
    """
    TANH OPERATOR:
    INPUT: REAL NUMBER OR AAD-VARIABLE
    RETURNS: AAD-VARIABLE TYPE
    """
    name = obj.name
    val = obj.val
    der = obj.der
    n_val = np.tanh(val)
    n_der = der * (1-(np.tanh(val)**2))
    return AADVariable(n_val,n_der,name=name)

def arcsin(obj: AADVariable) -> AADVariable:
    """
    ARCSIN OPERATOR:
    INPUT: REAL NUMBER OR AAD-VARIABLE
    RETURNS: AAD-VARIABLE TYPE
    """
    name = obj.name
    val = obj.val
    der = obj.der
    n_val = np.arcsin(val)
    n_der = der * 1/(np.sqrt(1-(val**2))) # 
    return AADVariable(n_val,n_der,name=name)

def arccos(obj: AADVariable) -> AADVariable:
    """
    ARCCOS OPERATOR:
    INPUT: REAL NUMBER OR AAD-VARIABLE
    RETURNS: AAD-VARIABLE TYPE
    """
    name = obj.name
    val = obj.val
    der = obj.der
    n_val = np.arccos(val)
    n_der = der * -1/(np.sqrt(1-(val**2))) #
    return AADVariable(n_val,n_der,name=name)

def arctan(obj: AADVariable) -> AADVariable:
    """
    ARCTAN OPERATOR:
    INPUT: REAL NUMBER OR AAD-VARIABLE
    RETURNS: AAD-VARIABLE TYPE
    """
    name = obj.name
    val = obj.val
    der = obj.der
    n_val = np.arctan(val)
    n_der = der * 1/(val**2+1) # 
    return AADVariable(n_val,n_der,name=name)

def sqrt(obj: AADVariable) -> AADVariable:
    """
    ARCTAN OPERATOR:
    INPUT: REAL NUMBER OR AAD-VARIABLE
    RETURNS: AAD-VARIABLE TYPE
    """
    name = obj.name
    val = obj.val
    der = obj.der
    n_val = np.sqrt(val)
    n_der = der * 0.5 * 1/(np.sqrt(val))
    return AADVariable(n_val,n_der,name=name)

def abs(obj: AADVariable) -> AADVariable:
    """
    ABSOLUTE VALUE OPERATOR:
    INPUT: REAL NUMBER OR AAD-VARIABLE
    RETURNS: AAD-VARIABLE TYPE
    """
    name = obj.name
    val = obj.val
    der = obj.der
    n_val = np.abs(val)
    n_der = der * val/np.abs(val)
    return AADVariable(n_val,n_der,name=name)



from AAD import AADVariable
import numpy as np

"""
Wrapper class for vector-valued functions in AAD.
"""
class AADFunction:
    """
    Initializer
    
    Example
    ----------

    Accepts input in AADVariables, like
    
    >>> x = AADVariable(1.0, [1.0])
    >>> y = AADVariable(2.0, [0.0, 1.0])
    >>> f = AADFunction([x*y, x+y])
    >>> print(f)
    [AADFunction fun = [2.0, 3.0], der = [[2. 1.]
     [1. 1.]]]
    """
    def __init__(self, fn):
        """
        Get the value of the vector-valued function.

        Input
        -----------
        fn      :function
                function to get the vector-valued function

        Example
        -----
        >>> x = AADVariable(1.0, [1.0])
        >>> y = AADVariable(2.0, [0.0, 1.0])
        >>> f = AADFunction([x*y, x+y])
        >>> print(f.val())
        [2.0, 3.0]
        """
        self.fn = fn
    

    def val(self):
        fn = self.fn
        if isinstance(fn, AADVariable):
            return self.fn.val
        elif isinstance(fn, np.ndarray) or isinstance(fn, list):
            return [v.val for v in self.fn]
        else: # const
            return self.fn


    def der(self):
        """
        Example
        -----------
        Get the Jacobian of the vector-valued function.
        >>> x = AADVariable(1.0, [1.0])
        >>> y = AADVariable(2.0, [0.0, 1.0])
        >>> f = AADFunction([x*y, x+y])
        >>> print(f.der())
        [[2. 1.]
         [1. 1.]]
        """
        fn = self.fn
        if isinstance(fn, AADVariable):
            return self.fn.der
        elif isinstance(fn, np.ndarray) or isinstance(fn, list):
            return np.array([v.der for v in self.fn])
        else: # const
            return 0
    
    def __str__(self):
        return "[AADFunction fun = " + self.val().__str__() + ", der = " + self.der().__str__() + "]"



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="aad-jimmielin", # Replace with your own username
    version="1.0.5",
    author="CS107-Dist-Computing-Group",
    author_email="hplin@seas.harvard.edu",
    description="Awesome Automatic Differentiation Project",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dist-computing/cs107-FinalProject",
    packages=setuptools.find_packages(),
    install_requires=["numpy", "scipy", "pytest"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
